// Copyright 2024 XMOS LIMITED.
// This Software is subject to the terms of the XMOS Public Licence: Version 1.

#include <print.h>

#include "i2s_in_wrapper.h"
#include "utils.h"
#include "app_config.h"
#include "i2s.h"

#if (CONFIG == I2S)

// Global written to by spdif_rx_custom and read by LED routine in utils
unsafe{
    unsigned i2s_rx_sr = 0;
    unsigned * unsafe i2s_rx_sr_ptr = &i2s_rx_sr;
}

[[distributable]]
void i2s_in_handler(chanend c_digital_rx_to_asrc, server i2s_frame_callback_if i_i2s){
    // Approx rate calculation based on sample counting
    timer tmr;
    int32_t rate_measurement_trigger = 0;
    const uint32_t measurement_rate_hz = 10;
    const int32_t rate_measurement_period = XS1_TIMER_HZ / measurement_rate_hz;
    tmr :> rate_measurement_trigger;
    rate_measurement_trigger += rate_measurement_period;

    // I2S sample rate measurement state
    uint32_t i2s_sample_period_count = 0;
    uint8_t measured_i2s_sample_rate_change = 1;    // Force new SR as measured
    uint32_t current_i2s_rate = 0;                  // Set to invalid until we measure it properly
    int32_t latest_timestamp = 0;
    int32_t last_timestamp = 0;

    while(1){
        select{
            case i_i2s.init(i2s_config_t &?i2s_config, tdm_config_t &?tdm_config):
                printstr("I2S slave input init: "); printintln(current_i2s_rate);
                if(i2s_config.slave_frame_synch_error){
                    printstrln("Restarted due to LR clk synch error");
                }
                i2s_config.slave_bclk_polarity = I2S_SLAVE_SAMPLE_ON_BCLK_RISING;
                i2s_config.mode = I2S_MODE_I2S;
            break;

            case i_i2s.restart_check() -> i2s_restart_t restart:
                // This if the first callback so the least jitter measurement of timestamp should be here
                tmr :> latest_timestamp;

                // Inform the I2S slave whether it should restart or exit
                i2s_sample_period_count++;
                if(measured_i2s_sample_rate_change){
                    measured_i2s_sample_rate_change = 0;
                    restart = I2S_RESTART;
                    break;
                }
                restart = I2S_NO_RESTART;
            break;

            case i_i2s.receive(size_t num_in, int32_t samples[num_in]):
#if APPLY_PRE_POST_GAIN
                // Do this before to reduce the ISR burden on ASRC
                for(int i = 0; i < num_in; i++){
                    samples[i] = small_attenuation(samples[i]);
                }
#endif
                // Send samples to ASRC
                outuint(c_digital_rx_to_asrc, current_i2s_rate);
                outuint(c_digital_rx_to_asrc, latest_timestamp);
                outuint(c_digital_rx_to_asrc, num_in);
                for(int i = 0; i < num_in; i++){
                    outuint(c_digital_rx_to_asrc, samples[i]);
                }

                // Calculate the nominal sample rate
                uint32_t measured_i2s_rate = calc_sample_rate(&last_timestamp, latest_timestamp, current_i2s_rate, &i2s_sample_period_count);

                if((measured_i2s_rate != 0) && (current_i2s_rate != measured_i2s_rate)){
                    measured_i2s_sample_rate_change = 1;
                    current_i2s_rate = measured_i2s_rate;
                    unsafe{ *i2s_rx_sr_ptr = current_i2s_rate;}

                }
            break;

            case i_i2s.send(size_t num_out, int32_t samples[num_out]):
                // Unused in this application. This case does not fire.
            break;
        } // select
    }
}

void i2s_rx_task(chanend c_digital_rx_to_asrc,
                in port p_bclk_in,
                clock bclk_in,
                buffered in port:32 p_lrclk_in,
                buffered in port:32 p_adc_in[]){

    i2s_frame_callback_if i_i2s_in;


    par {
        [[distribute]] // This causes audio_hub to be a callback from i2s_frame_slave and not consume a thread
        i2s_in_handler(c_digital_rx_to_asrc, i_i2s_in);
        i2s_frame_slave(i_i2s_in, NULL, NUM_I2S_IN_DAC_LINES, p_adc_in, NUM_I2S_IN_ADC_LINES, I2S_DATA_BITS, p_bclk_in, p_lrclk_in, bclk_in);
    }
}

#endif // CONFIG == I2S
