// Copyright 2024 XMOS LIMITED.
// This Software is subject to the terms of the XMOS Public Licence: Version 1.

#include <platform.h>
#include <xs1.h>
#include <stdint.h>
#include <string.h>
#include <print.h>
#include "xassert.h"
#ifdef ASYNC_FIFO_XSCOPE_INSTRUMENTATION
#include <xscope.h>
#endif

#include "app_config.h"

#include "i2s.h"
#include "i2c.h"
#include "utils.h"
extern "C"{
    #include "asrc_task.h"
    #include "sw_pll.h"
}

extern void board_setup(void);

#if (CONFIG == ADAT)
// Digital Tx/Rx resources
on tile[0]: buffered in port:32 p_dig_in =                  PORT_ADAT_IN;   // Optical connector
on tile[1]: buffered out port:32 p_dig_out =                PORT_ADAT_OUT;  // Optical connector
on tile[1]: clock dig_tx_blk =                              XS1_CLKBLK_2;
#elif (CONFIG == SPDIF)
on tile[0]: in port p_dig_in =                              PORT_ADAT_IN;   // Optical connector
on tile[1]: buffered out port:32 p_dig_out =                PORT_ADAT_OUT;  // Optical connector
on tile[0]: clock spdif_rx_clk =                            XS1_CLKBLK_1;
on tile[1]: clock dig_tx_blk =                              XS1_CLKBLK_2;
#elif (CONFIG == I2S)
on tile[0]: in port p_bclk_in =                             XS1_PORT_1A;    // PLL_SYNC_OUT on J34
on tile[0]: buffered in port:32  p_lrclk_in =               XS1_PORT_1P;    // WORD_CLOCK_IN on J34
on tile[0]: clock bclk_in =                                 XS1_CLKBLK_1;
on tile[0]: buffered in port:32  p_adc_in[] =               {XS1_PORT_1K};  // X0D34 on J34
#endif

// I2C resources (defined in lib_board_support)
extern port p_scl;
extern port p_sda;

// GPIO resources
on tile[0]: in port p_buttons =                             XS1_PORT_4E;
on tile[0]: out port p_leds =                               XS1_PORT_4F;

// I2S resources
on tile[1]: in port p_mclk =                                PORT_MCLK_IN;
on tile[1]: buffered in port:32 p_lrclk =                   PORT_I2S_LRCLK;
on tile[1]: in port p_bclk =                                PORT_I2S_BCLK;
#if (CONFIG == ADAT)
on tile[1]: buffered out port:32 p_dac[NUM_I2S_DAC_LINES] = {PORT_I2S_DAC0, PORT_I2S_DAC1, PORT_I2S_DAC2, PORT_I2S_DAC3};
on tile[1]: buffered in port:32 p_adc[NUM_I2S_ADC_LINES] =  {PORT_I2S_ADC0 ,PORT_I2S_ADC1, PORT_I2S_ADC2, PORT_I2S_ADC3};
#elif (CONFIG == SPDIF)
on tile[1]: buffered out port:32 p_dac[NUM_I2S_DAC_LINES] = {PORT_I2S_DAC0};
on tile[1]: buffered in port:32 p_adc[NUM_I2S_ADC_LINES] =  {PORT_I2S_ADC0};
#elif (CONFIG == I2S)
on tile[1]: buffered out port:32 p_dac[NUM_I2S_DAC_LINES] = {PORT_I2S_DAC0};
#define p_adc                                               NULL
#endif
on tile[1]: clock bclk =                                    XS1_CLKBLK_1;


[[distributable]]
void audio_hub( chanend ?c_dig_tx,
                server i2s_frame_callback_if i_i2s,
                asrc_in_out_t * unsafe asrc_io,
                asynchronous_fifo_t * unsafe fifo_ptr) {
    // Approx rate calculation based on sample counting
    timer tmr;
    int32_t rate_measurement_trigger = 0;
    const uint32_t measurement_rate_hz = 10;
    const int32_t rate_measurement_period = XS1_TIMER_HZ / measurement_rate_hz;
    tmr :> rate_measurement_trigger;
    rate_measurement_trigger += rate_measurement_period;

    // I2S sample rate measurement state
    uint32_t i2s_sample_period_count = 0;
    uint8_t measured_i2s_sample_rate_change = 1;    // Force new SR as measured
    uint32_t current_i2s_rate = 0;                  // Set to invalid until we measure it properly

    // General control
    uint32_t mute = 0;
    int32_t latest_timestamp = 0;
    int32_t last_timestamp = 0;

#if (CONFIG == ADAT)
    // ADAT Tx
    printstr("ADAT Rx TO I2S SLAVE\n");
    int adat_tx_smux = SMUX_NONE;
    int32_t adat_tx_samples[ADAT_MAX_SAMPLES] = {0};
    adat_tx_smux = adat_tx_startup(c_dig_tx, current_i2s_rate, adat_tx_samples);
#elif (CONFIG == SPDIF)
    printstr("SPDIF Rx TO I2S SLAVE\n");
    spdif_tx_startup(c_dig_tx, current_i2s_rate, get_master_clock_from_samp_rate(current_i2s_rate));
#elif (CONFIG == I2S)
    printstr("I2S SLAVE Rx TO I2S SLAVE\n");
#endif

    while(1) {
        select{
            case i_i2s.init(i2s_config_t &?i2s_config, tdm_config_t &?tdm_config):
                printstr("I2S slave output init: "); printintln(current_i2s_rate);
                i2s_config.slave_bclk_polarity = I2S_SLAVE_SAMPLE_ON_BCLK_RISING; // Has no effect when master
                i2s_config.mode = I2S_MODE_I2S;

#if (CONFIG == ADAT)
                adat_tx_shutdown(c_dig_tx);
#elif (CONFIG == SPDIF)
                spdif_tx_shutdown(c_dig_tx);
#endif
                mute = (current_i2s_rate * FORMAT_CHANGE_MUTE_MS) / 1000;
                i2s_config.mode = I2S_MODE_I2S;

                reset_asrc_fifo_consumer(fifo_ptr);

#if (CONFIG == ADAT)
                adat_tx_smux = adat_tx_startup(c_dig_tx, current_i2s_rate, adat_tx_samples);
#elif (CONFIG == SPDIF)
                spdif_tx_startup(c_dig_tx, current_i2s_rate, get_master_clock_from_samp_rate(current_i2s_rate));
#endif
            break;

            case i_i2s.restart_check() -> i2s_restart_t restart:
                // This if the first callback so the least jitter measurement of timestamp should be here
                tmr :> latest_timestamp;

                // Inform the I2S slave whether it should restart or exit
                i2s_sample_period_count++;
                if(measured_i2s_sample_rate_change){
                    measured_i2s_sample_rate_change = 0;
                    restart = I2S_RESTART;
                    break;
                }
                restart = I2S_NO_RESTART;
            break;

            case i_i2s.receive(size_t num_in, int32_t samples[num_in]):
                if(mute == 0){
#if USE_DAC_AS_I2S_MASTER
#if (CONFIG == ADAT)
                    memcpy(adat_tx_samples, samples, num_in * sizeof(int32_t));
                    send_adat_tx_samples(c_dig_tx, (unsigned *)adat_tx_samples, adat_tx_smux);
#elif (CONFIG == SPDIF)
                    spdif_tx_output(c_dig_tx, samples[0], samples[1]);
#endif
#endif
                }
            break;

            case i_i2s.send(size_t num_out, int32_t samples[num_out]):
                int32_t asrc_out[MAX_ASRC_CHANNELS_TOTAL];
                int asrc_channel_count = pull_samples(asrc_io, fifo_ptr, asrc_out, current_i2s_rate, latest_timestamp);
                if(mute > 0 || asrc_channel_count == 0){
                    for(int ch = 0; ch < num_out; ch++){
                        samples[ch] = 0;
                    }
                }
                else {
                    for(int ch = 0; ch < num_out; ch++){
#if APPLY_PRE_POST_GAIN
                        samples[ch] = small_amplification(asrc_out[ch]);
#else
                        samples[ch] = asrc_out[ch];
#endif
                    }
                }
                if(mute){
                    mute--;
                }
                uint32_t measured_i2s_rate = calc_sample_rate(&last_timestamp, latest_timestamp, current_i2s_rate, &i2s_sample_period_count);

                if((measured_i2s_rate != 0) && (current_i2s_rate != measured_i2s_rate)){
                    measured_i2s_sample_rate_change = 1;
                    current_i2s_rate = measured_i2s_rate;
                }
            break;
        } // select
    }// while(1)
}


extern void init_asrc_io_callback_wrapper(asrc_in_out_t * unsafe asrc_io);
int main(void) {
    // Cross tile channels
    chan c_digital_rx_to_asrc;

    par {
        on tile[0]: {
            // Local to tile channels and interfaces
            interface i2c_master_if i2c[1];
            c_smux_change_adat_rx_declaration; // This is only used for ADAT
            streaming chan c_adat_rx;

            xk_audio_316_mc_ab_board_setup(hw_config);

            par{
#if (CONFIG == ADAT)
                adat_rx_task(c_adat_rx, p_dig_in);
                adat_rx_demux(c_adat_rx, c_digital_rx_to_asrc, c_smux_change_adat_rx_name);
#elif (CONFIG == SPDIF)
                spdif_rx_task(c_digital_rx_to_asrc, p_dig_in, spdif_rx_clk);
#elif (CONFIG == I2S)
                i2s_rx_task(c_digital_rx_to_asrc, p_bclk_in, bclk_in, p_lrclk_in, p_adc_in);
#endif
                // These can be deleted when deployed in custom hardware. Only for demo.
                gpio(c_smux_change_adat_rx_name, p_buttons, p_leds, i2c[0]);
                [[distribute]] // This causes i2c to be called inline rather than a separate thread
                i2c_master(i2c, 1, p_scl, p_sda, 100);
            }
        } // tile 0
        on tile[1]: unsafe{
            // Local to tile channels and interfaces
            i2s_frame_callback_if i_i2s;
            chan_c_dig_tx_declaration;

            // Shared memory items on tile[1]
            // IO struct for ASRC must be passed to both asrc_proc and consumer
            asrc_in_out_t asrc_io = {{{0}}};
            asrc_in_out_t * unsafe asrc_io_ptr = &asrc_io;

            // FIFO and ASRC I/O declaration. Global to allow producer and consumer to access it
            int64_t array[ASYNCHRONOUS_FIFO_INT64_ELEMENTS(MAX_FIFO_LENGTH, MAX_ASRC_CHANNELS_TOTAL)] = {0};
            asynchronous_fifo_t * unsafe fifo_ptr = (asynchronous_fifo_t *)array;

            init_asrc_io_callback_wrapper(&asrc_io);
#ifdef ASYNC_FIFO_XSCOPE_INSTRUMENTATION
            xscope_mode_lossy();
#endif

            par {
                [[distribute]] // This causes audio_hub to be a callback from i2s_frame_slave and not consume a thread
                audio_hub(chan_c_dig_tx_name, i_i2s, asrc_io_ptr, fifo_ptr);
                i2s_frame_slave(i_i2s, p_dac, NUM_I2S_DAC_LINES, p_adc, NUM_I2S_ADC_LINES, I2S_DATA_BITS, p_bclk, p_lrclk, bclk);
                asrc_task(c_digital_rx_to_asrc, asrc_io_ptr, fifo_ptr, MAX_FIFO_LENGTH);   // Spawns multiple threads
#if (CONFIG == ADAT)
                adat_tx_task(chan_c_dig_tx_name, p_dig_out, dig_tx_blk, p_mclk);
#elif (CONFIG == SPDIF)
                spdif_tx_task(chan_c_dig_tx_name, p_dig_out, dig_tx_blk, p_mclk);
#endif
            }
        } // tile 1
    } // par

    return 0;
}
