// Copyright 2024 XMOS LIMITED.
// This Software is subject to the terms of the XMOS Public Licence: Version 1.

#ifndef _UTILS_H_
#define _UTILS_H_

#include <platform.h>
#include <xs1.h>
#include <stdint.h>

#include "app_config.h"
#include "i2c.h"

enum audio_port_idx{
    IO_I2S = 0,
    IO_ADAT_RX,
    IO_ADAT_TX
};

// These are statically calculated by the compiler for speed
#define RATE_LOWER(rate, ppm) ((uint32_t)((float)rate * (1.0 - (float)ppm / 1e6) + 0.5))
#define RATE_HIGHER(rate, ppm) ((uint32_t)((float)rate * (1.0 + (float)ppm / 1e6)+ 0.5))
#define CHECK_RATE(sps, rate, ppm) if (sps > RATE_LOWER(rate, ppm) && sps < RATE_HIGHER(rate, ppm)){return rate;}

// This function takes an approximate sample rate +-  tolerance and snaps it to a normal sample rate
inline uint32_t get_normal_sample_rate(uint32_t samples_per_second){
    const uint32_t ppm_tolerance = 10000;

    // Check in reverse order so high rates take fewer cycles
    CHECK_RATE(samples_per_second, 192000, ppm_tolerance);
    CHECK_RATE(samples_per_second, 176400, ppm_tolerance);
    CHECK_RATE(samples_per_second, 96000, ppm_tolerance);
    CHECK_RATE(samples_per_second, 88200, ppm_tolerance);
    CHECK_RATE(samples_per_second, 48000, ppm_tolerance);
    CHECK_RATE(samples_per_second, 44100, ppm_tolerance);

    return 0; // Invalid rate found
}


// Calculates the expected MCLK from a given normal sample rate
inline uint32_t get_master_clock_from_samp_rate(uint32_t sample_rate){
    return (sample_rate % 48000 == 0) ? MCLK_48 : MCLK_441;
}


// Returns the normal sample rate calculate by measuring number of sample over a period defined by I2S_RATE_MEASUREMENT_HZ
inline uint32_t calc_sample_rate(int32_t *last_timestamp, int32_t latest_timestamp, uint32_t current_i2s_rate, uint32_t *i2s_sample_period_count){
    const int32_t rate_measurement_period = XS1_TIMER_HZ / I2S_RATE_MEASUREMENT_HZ;

    if(timeafter(latest_timestamp, *last_timestamp + rate_measurement_period)){
        *last_timestamp = latest_timestamp;
        uint32_t samples_per_second = *i2s_sample_period_count * I2S_RATE_MEASUREMENT_HZ;
        *i2s_sample_period_count = 0;

        return get_normal_sample_rate(samples_per_second);
    }

    return current_i2s_rate;
}


// Experimental. Uses timestamps to calculate sample rate
inline uint32_t sample_rate_from_ts_diff(int32_t t0, int32_t t1){
    int32_t ts_diff = t1 - t0;
    if(ts_diff == 0){
        return 0;
    }
    uint32_t approx_rate = XS1_TIMER_HZ / ts_diff;

    return get_normal_sample_rate(approx_rate);
}

// GPIO task. Handles buttons and leds and connects to I2C master to allow configuration of on board hardware for given sample-rate.
void gpio(chanend ?c_smux_change_adat_rx, in port p_buttons, out port p_leds, client interface i2c_master_if i_i2c);


// High quality attenuate and amplify by a small amount to avoid clipping in
// the ASRC filter banks when sharp input signals caused by the Gibbs phenomenon
// Attenuate can be applied pre-ASRC and gain post-ASRC. Default factor set to
// 0.89 which is 1 dB
// https://en.wikipedia.org/wiki/Gibbs_phenomenon

#define GAIN_SHIFT      30      // Bit shift for multiplier
#define GAIN_FACTOR     0.89f    // Gain/attenuate factor. This works out to around 1 dB

static inline int32_t small_attenuation(int32_t val){
    int32_t mul = (int32_t)(GAIN_FACTOR * (1 << GAIN_SHIFT));
    int64_t tmp = (int64_t)mul * (int64_t)val + (1 << (GAIN_SHIFT - 1));
    return (int32_t)(tmp >> GAIN_SHIFT); 
}

static inline int32_t small_amplification(int32_t val){
    int32_t mul = (int32_t)((1 << GAIN_SHIFT) / GAIN_FACTOR);
    int64_t tmp = (int64_t)mul * (int64_t)val + (1 << (GAIN_SHIFT - 1));
    return (int32_t)(tmp >> GAIN_SHIFT); 
}

#endif // _UTILS_H_
