// Copyright 2024 XMOS LIMITED.
// This Software is subject to the terms of the XMOS Public Licence: Version 1.

#include <platform.h>
#include <xs1.h>
#include <stdint.h>
#include <stdio.h>

#include "app_config.h"
#include "adat_wrapper.h"
#include "utils.h"


typedef struct button_state_t{
    uint8_t active_level;   // gpio level for active
    uint8_t counter;        // counts down for debounce
    uint8_t bit_shift;      // Which bit to monitor
}button_state_t;

const unsigned sr_list[] = {44100, 48000, 88200, 96000, 176400, 192000};
const unsigned smux_list[] = {SMUX_NONE, SMUX_II, SMUX_IV};


unsigned get_format_idx_from_sr(unsigned sr){
    for(int i = 0; i < sizeof(sr_list) / sizeof(sr_list[0]); i++){
        if(sr == sr_list[i]){
            return i;
        }
    }
    return 0;
}


static unsigned curr_sr_idx = 0;

unsigned button_action(client interface i2c_master_if i_i2c, chanend ?c_smux_change_adat_rx, int idx){
    if(idx == 2){
        curr_sr_idx++;
        if(curr_sr_idx == sizeof(sr_list) / sizeof(sr_list[0])){
            curr_sr_idx = 0;
        }
        // Set new sample rate on CODEC (I2S master)
        xk_audio_316_mc_ab_AudioHwConfig(i_i2c, hw_config, sr_list[curr_sr_idx], get_master_clock_from_samp_rate(sr_list[curr_sr_idx]), 0, 24, 24);

        printf("Configured I2S master to: %u\n", sr_list[curr_sr_idx]);
    }
#if (CONFIG == ADAT)
    static unsigned rx_smux_iv_flag = 0; // When set, presence of the SMUX bit will mean SMUX-IV instead of SMUX-II

    if(idx == 1){
        rx_smux_iv_flag ^= 1; // Toggle SMUX IV flag

        // Send to ADAT Rx demux
        c_smux_change_adat_rx <: IO_ADAT_RX;
        c_smux_change_adat_rx <: rx_smux_iv_flag;
    }
#endif
    if(idx == 0){
        // Do nothing. Not currently used.
    }

    return curr_sr_idx;
}

// Shared memory vars for communicating the input sample_rate
extern unsigned * unsafe spdif_rx_sr_ptr;
extern uint32_t * unsafe adat_rx_sr_ptr;

#define NUM_BUTTONS 3
void gpio(chanend ?c_smux_change_adat_rx, in port p_buttons, out port p_leds, client interface i2c_master_if i_i2c){

    // Initial audio HW setup
    xk_audio_316_mc_ab_AudioHwInit(i_i2c, hw_config);
    xk_audio_316_mc_ab_AudioHwConfig(i_i2c, hw_config, DEFAULT_FREQ, get_master_clock_from_samp_rate(DEFAULT_FREQ), 0, 24, 24);

    // Buttons
    const uint8_t counts_for_active = 20;
    const uint8_t active_level = 0;
    button_state_t button_state[NUM_BUTTONS] =  {{active_level, counts_for_active, 0},
                                                 {active_level, counts_for_active, 1},
                                                 {active_level, counts_for_active, 2}};
    timer tmr_buttons;
    int32_t sample_time;
    tmr_buttons :> sample_time;


    // LEDs
    const unsigned led_sr = 0x1;    // I2S sample rate
    const unsigned led_dr = 0x8;    // Digital Rx state

    timer tmr_leds;
    int32_t led_transition_time;
    tmr_leds :> led_transition_time;
    unsigned led_val = 0; // all off
    unsigned led_sequence_idx = 0;

    // State
    unsigned sr_i2s_idx = 0; // This will be overwritten by next function
    // Note in the case of SPDIF Rx, the flash index is determined in the LED select case below.


    // Reverse lookup DEFAULT_FREQ
    for(int i = 0; i < sizeof(sr_list) / sizeof(sr_list[0]); i++){
        if(sr_list[i] == DEFAULT_FREQ) curr_sr_idx = i;
    }

    while (1) {
        select {
            // wait til the timer reaches the timeout to enter a stable period (debounce)
            case tmr_buttons when timerafter(sample_time) :> void:
                uint32_t port_val;
                p_buttons :> port_val;
                for(int i = 0; i < NUM_BUTTONS; i++){
                    uint8_t mask = 1 << button_state[i].bit_shift;
                    if((port_val & mask) == (button_state[i].active_level << button_state[i].bit_shift)){
                        if(button_state[i].counter){
                            button_state[i].counter--;
                            if(button_state[i].counter == 0){
                                sr_i2s_idx = button_action(i_i2c, c_smux_change_adat_rx, i);
                            }
                        }
                    } else {
                        button_state[i].counter = counts_for_active;
                    }
                }
            sample_time += XS1_TIMER_KHZ * 1;
            break;

            // Flash LEDs according to SR
            case tmr_leds when timerafter(led_transition_time) :> void:
                led_transition_time += XS1_TIMER_KHZ * 150;

                if(led_sequence_idx < 2 * (sr_i2s_idx+1)){
                    if(led_sequence_idx % 2 == 0) led_val |= led_sr;
                    else led_val &= ~led_sr;
                }

                unsigned dig_rx_sr_idx = 0;
#if (CONFIG == ADAT)
                unsafe{dig_rx_sr_idx = get_format_idx_from_sr(*adat_rx_sr_ptr);}

#elif (CONFIG == SPDIF)
                unsafe{dig_rx_sr_idx = get_format_idx_from_sr(*spdif_rx_sr_ptr);}
#endif
                if(led_sequence_idx < 2 * (dig_rx_sr_idx+1)){
                    if(led_sequence_idx % 2 == 0) led_val |= led_dr;
                    else led_val &= ~led_dr;
                }

                led_sequence_idx++;
                if (led_sequence_idx > 2 * 7){
                    led_sequence_idx = 0;
                }

                p_leds <: led_val;
            break;
        }
    }
}
